%{
#include "inca.h"
#include "inca.parser.h"

int lineno = 1;
%}

%%
minimize		yylval.string=strdup(yytext); return OPTIMIZATION;
maximize		yylval.string=strdup(yytext); return OPTIMIZATION;
abs			yylval.string=strdup(yytext); return ABS;
[0-9]+			yylval.number=  atoi(yytext); return INTEGER;
[a-z_][a-zA-Z0-9_]*	yylval.string=strdup(yytext); return SYMBOL;
\$[a-z_][a-zA-Z0-9_]*	yylval.string=strdup(yytext); return CONSTRAINTVARIABLE;
\$\$[a-z_][a-zA-Z0-9_]*	yylval.string=strdup(yytext); return CONSTRAINT;
[A-Z][a-zA-Z0-9_]*	yylval.string=strdup(yytext); return VARIABLE;
,			return COMMA;
:-			return NECK;
\(			return LPAR;
\)			return RPAR;
\[			return LBRACK;
\]			return RBRACK;
\{			return LSET;
\}			return RSET;
not[ ]			return DNEG;
\.			return END;
\.\.			return RANGE;
:			return CONDITIONAL;
;			return POOLING;
=			return ASSIGNMENT;
#var[ ]			return DECLARATION;
#hide[^\n]*\n		++lineno; ECHO;
#show[^\n]*\n		++lineno; ECHO;
#const[^\n]*\n		++lineno; ECHO;
[ \t]+			/* whitespaces */
\%[^\n]*		/* comment */
\n			++lineno;
\+			yylval.string=strdup(yytext); return ADD;
\-			yylval.string=strdup(yytext); return SUB;
\*			yylval.string=strdup(yytext); return MUL;
\/			yylval.string=strdup(yytext); return DIV;
[=\!\<\>]=		yylval.string=strdup(yytext); return COMPARISON;
[\<\>]			yylval.string=strdup(yytext); return COMPARISON;
.			return CHARACTER;
%%

void yyerror(const char *str) {
	fprintf(stderr, "error in line %d: %s\n", lineno, str);
}

int yywrap() {
	return 1;
}

