%{
#include "inca.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

rule* createRul(pred*, pred*);
pred* createSop(char*, pred*);
pred* createWop(char*, pred*);
pred* createSet(pred*, pred*);
pred* createBod(pred*, pred*);
pred* createClt(pred*, pred*);
pred* createWlt(pred*, term*);
pred* createACC(term*, pred*, term*);
pred* createAWC(term*, pred*, term*);
pred* createCpr(char*, term*, term*);
pred* createCon(char*, term*, term*);
pred* createAss(char*, term*);
pred* createAtm(char*, term*);
term* createCVr(char*, term*);
term* createAlt(term*, term*);
term* createRng(term*, term*);
term* createArg(term*, term*);
term* createCcd(term*, pred*);
term* createCcv(term*, pred*);
term* createPar(term*);
term* createNeg(term*);
term* createExp(char*, term*, term*);
term* createFun(char*, term*);
term* createVar(char*);
term* createInt(int);
term* createNil();

void free_rule(rule*);
void free_pred(pred*);
void free_term(term*);
%}

%union {
	int number;
	char *string;
	term *term;
	pred *pred;
	rule *rule;
}

%token <number> INTEGER
%token <string> CONSTRAINT CONSTRAINTVARIABLE SYMBOL VARIABLE ADD SUB MUL DIV COMPARISON OPTIMIZATION ABS
%token COMMA NECK LPAR RPAR LBRACK RBRACK LSET RSET DNEG END RANGE CONDITIONAL POOLING ASSIGNMENT DECLARATION CHARACTER

%type <term> cset ccond pool args term expr cvar
%type <pred> opt con head body blit agg set list wlit cond clit comp dlit atom
%type <rule> rule

%right POOLING CONDITIONAL

%left ADD SUB
%left MUL DIV

%destructor { free($$); } CONSTRAINT CONSTRAINTVARIABLE SYMBOL VARIABLE ADD SUB MUL DIV COMPARISON OPTIMIZATION ABS
%destructor { free_term($$); } cset ccond pool args term expr cvar
%destructor { free_pred($$); } opt con head body blit agg set list wlit cond clit comp dlit atom
%destructor { free_rule($$); } rule

%error-verbose
%%
prog:	| prog stmt ;

stmt:	rule { print_rule($1); free_rule($1) } |
	DECLARATION cset ASSIGNMENT pool END { int errs = 0; enc_vars($2, $4);
		if ($4->cvar || $4->vars || $4->type == TTVAR || $4->type == TTCVR) {
			yyerror("unrestricted domain");
			errs++;
		}
		if (v_enc != VDIRECT && $4->alt) {
			yyerror("pooling used under range or bound encoding");
			errs++;
		}
		free_term($2); free_term($4);
		if (errs) YYERROR;
	} |
	opt { print_pred($1); free_pred($1); printf(".\n") } ;

opt:	OPTIMIZATION LSET set RSET END      { $$ = createSop($1, $3) } |
	OPTIMIZATION LBRACK list RBRACK END { $$ = createWop($1, $3) } ;

con:	CONSTRAINT LSET cset RSET                                { $$ = createCon($1, $3, NULL) } |
	CONSTRAINT LPAR pool RPAR LSET cset RSET                 { $$ = createCon($1, $6, $3) } |
	CONSTRAINT LSET cset RSET ASSIGNMENT term                { $$ = createCon($1, $3, $6) } |
	CONSTRAINT LPAR pool RPAR LSET cset RSET ASSIGNMENT term { $$ = createCon($1, $6, createArg($9, $3)) } ;

cset:	ccond COMMA cset { $$ = createArg($1, $3) } |
	ccond            { $$ = $1 } ;

ccond:	cvar CONDITIONAL clit { $$ = createCcv($1, $3) } |
	cvar                  { $$ = $1 } ;

rule:	head END             { $$ = createRul($1  , NULL) } |
	head NECK body END   { $$ = createRul($1  ,   $3) } |
	     NECK body END   { $$ = createRul(NULL,   $2) } ;

head:	agg  { $$ = $1 } |
	con  { $$ = $1 } |
	atom { $$ = $1 } |
	expr COMPARISON expr { $$ = createCpr($2, $1, $3) } ;

body:	blit COMMA body { $$ = createBod($1, $3) } |
	blit            { $$ = $1 } ;

blit:	cond     { $$ = $1 } |
	comp     { $$ = $1 } |
	DNEG agg { $$ = $2; $$->sign = 1 } |
	agg      { $$ = $1 } |
	DNEG con { $$ = $2; $$->sign ^= 1 } |
	con      { $$ = $1 } ;

agg:	expr LSET set RSET expr      { $$ = createACC($1, $3, $5) } |
	expr LSET set RSET           { $$ = createACC($1, $3, createNil()) } |
	     LSET set RSET expr      { $$ = createACC(createNil(), $2, $4) } |
	     LSET set RSET           { $$ = createACC(createNil(), $2, createNil()) } |
	expr LBRACK list RBRACK expr { $$ = createAWC($1, $3, $5) } |
	expr LBRACK list RBRACK      { $$ = createAWC($1, $3, createNil()) } |
	     LBRACK list RBRACK expr { $$ = createAWC(createNil(), $2, $4) } |
	     LBRACK list RBRACK      { $$ = createAWC(createNil(), $2, createNil()) } ;

set:	cond COMMA set { $$ = createSet($1, $3) } |
	cond           { $$ = $1 } ;

list:	wlit COMMA list { $$ = createSet($1, $3) } |
	wlit            { $$ = $1 } ;

wlit:	cond ASSIGNMENT expr { $$ = createWlt($1, $3) } |
	cond                 { $$ = createWlt($1, createInt(1)) } ;

cond:	dlit CONDITIONAL clit { $$ = createClt($1, $3) } |
	dlit                  { $$ = $1 } ;

clit:	clit CONDITIONAL clit { $$ = createClt($1, $3) } |
	atom                  { $$ = $1 } |
	comp                  { $$ = $1 } ;

comp:	VARIABLE ASSIGNMENT expr { $$ = createAss($1, $3) } |
	expr COMPARISON expr     { $$ = createCpr($2, $1, $3) } ;

dlit:	DNEG atom { $$ = $2; $$->sign = 1 } |
	atom      { $$ = $1 } ;

atom:	SYMBOL                { $$ = createAtm($1, NULL) } |
	SYMBOL LPAR pool RPAR { $$ = createAtm($1, $3) } ;

pool:	pool POOLING pool  { $$ = createAlt($1, $3) } |
	expr RANGE expr { $$ = createRng($1, $3) } |
	args            { $$ = $1 } ;

args:	term COMMA args { $$ = createArg($1, $3) } |
	term            { $$ = $1 } ;

term:	SYMBOL LPAR args RPAR { $$ = createFun($1, $3) } |
	expr                  { $$ = $1 } ;

expr:	expr ADD expr  { $$ = createExp($2, $1, $3) } |
	expr SUB expr  { $$ = createExp($2, $1, $3) } |
	expr MUL expr  { $$ = createExp($2, $1, $3) } |
	expr DIV expr  { $$ = createExp($2, $1, $3) } |
	LPAR expr RPAR { $$ = createPar($2) } |
	SUB expr       { $$ = createNeg($2); free($1) } |
	SYMBOL         { $$ = createFun($1, NULL) } |
	VARIABLE       { $$ = createVar($1) } |
	INTEGER        { $$ = createInt($1) } |
	ABS LPAR expr RPAR { $$ = createFun($1, $3) } |
	cvar           { $$ = $1 } ;

cvar:	CONSTRAINTVARIABLE                { $$ = createCVr($1, NULL) } |
	CONSTRAINTVARIABLE LPAR args RPAR { $$ = createCVr($1, $3) } ;

%%
int num_cons = 0;
int num_vars = 0;

term* appendVars(term* a, term* b) {
	if (!a) return b;
	if (!b) return a;
	term* t = a;
	while (t->vars)
		t = t->vars;
	t->vars = b;
	return a;
}

term* appendCvar(term* a, term* b) {
	if (!a) return b;
	if (!b) return a;
	term* t = a;
	while (t->cvar)
		t = t->cvar;
	t->cvar = b;
	return a;
}

term* createTerm(t_term type, char *string, int number, struct linked_term *args) {
	term *tmp = (term*)malloc(sizeof(term));
	tmp->type = type;
	tmp->string = string;
	tmp->number = number;
	tmp->arity = args ? args->index + 1 : 0;
	tmp->index = 0;
	tmp->ident = 0;
	tmp->args = args;
	tmp->alt = NULL;
	tmp->next = NULL;
	tmp->vars = NULL;
	tmp->cvar = NULL;
	term* arg = args;
	while (arg) {
		if (arg->type == TTVAR)
			tmp->vars = appendVars(tmp->vars, arg);
		if (arg->type == TTCVR)
			tmp->cvar = appendCvar(tmp->cvar, arg);
		if (arg->type != TTVAR)
			tmp->vars = appendVars(tmp->vars, arg->vars);
		if (arg->type != TTCVR)
			tmp->cvar = appendCvar(tmp->cvar, arg->cvar);
		arg = arg->next;
	}
	tmp->body = NULL;
	return tmp;
}

term* createAlt(term* t, term* alt) {
	t->alt = alt;
	if (alt->type == TTVAR)
		t->vars = appendVars(t->vars, alt);
	if (alt->type == TTCVR)
		t->cvar = appendCvar(t->cvar, alt);
	if (alt->type != TTVAR)
		t->vars = appendVars(t->vars, alt->vars);
	if (alt->type != TTCVR)
		t->cvar = appendCvar(t->cvar, alt->cvar);
	return t;
}

term* createCcv(term* v, pred* c) {
	v->body = c;
	v->vars = appendVars(v->vars, c->vars);
	return v;
}

term* createCVr(char* str, term* args) {
	term* tmp = createTerm(TTCVR, strdup(str+1), 0, args);
	tmp->ident = num_vars++;
	free(str);
	return tmp;
}

term* createRng(term* lower, term* upper) {
	return createTerm(TTRNG, NULL, 0, createArg(lower, upper));
}

term* createArg(term* head, term* tail) {
	head->next = tail;
	head->index = tail->index + 1;
	return head;
}

term* createPar(term* arg) {
	return createTerm(TTPAR, NULL, 0, arg);
}

term* createNeg(term* arg) {
	return createTerm(TTNEG, NULL, 0, arg);
}

term* createExp(char* str, term* arg1, term* arg2) {
	return createTerm(TTEXP, str, 0, createArg(arg1, arg2));
}

term* createFun(char* str, term* args) {
	return createTerm(TTFUN, str, 0, args);
}

term* createVar(char* str) {
	return createTerm(TTVAR, str, 0, NULL);
}

term* createInt(int i) {
	return createTerm(TTINT, NULL, i, NULL);
}

term* createNil() {
	return createTerm(TTFUN, strdup(""), 0, NULL);
}

pred* createPred(t_pred type, char *string, term* args, pred* body) {
	pred *tmp = (pred*)malloc(sizeof(pred));
	tmp->type = type;
	tmp->string = string;
	tmp->arity = args ? args->index + 1 : 0;
	tmp->index = 0;
	tmp->sign = 0;
	tmp->ident = 0;
	tmp->args = args;
	tmp->weight = NULL;
	tmp->vars = NULL;
	tmp->cvar = NULL;
	term* arg = args;
	while (arg) {
		if (arg->type == TTVAR)
			tmp->vars = appendVars(tmp->vars, arg);
		else if (arg->type == TTCVR)
			tmp->cvar = appendCvar(tmp->cvar, arg);
		if (arg->type != TTVAR)
			tmp->vars = appendVars(tmp->vars, arg->vars);
		if (arg->type != TTCVR)
			tmp->cvar = appendCvar(tmp->cvar, arg->cvar);
		arg = arg->next;
	}
	tmp->lits = NULL;
	tmp->body = body;
	tmp->next = NULL;
	tmp->pbody = NULL;
	return tmp;
}

pred* createCon(char* string, term* args, term* params) {
	pred* tmp = createPred(PTCON, strdup(string+2), params, NULL);
	tmp->lits = args;
	tmp->sign = 1;
	tmp->ident = num_cons++;
	free(string);
	term* it = args;
	while (it) {
		tmp->vars = appendVars(tmp->vars, it->vars);
		it = it->next;
	}
	return tmp;
}

pred* createSet(pred* head, pred* tail) {
	if (!tail) return head;
	head->next = tail;
	head->index = tail->index + 1;
	return head;
}

pred* createBod(pred* head, pred* tail) {
	if (!tail) return head;
	head->next = tail;
	head->index = tail->index + 1;
	if (!tail->sign && tail->type == PTREG)
		head->pbody = tail;
	else
		head->pbody = tail->pbody;
	return head;
}

pred* createSop(char* string, pred* lits) {
	return createPred(PTACC, string, createArg(createNil(), createNil()), lits);
}

pred* createWop(char* string, pred* lits) {
	return createPred(PTAWC, string, createArg(createNil(), createNil()), lits);
}

pred* createACC(term* lower, pred* body, term* upper) {
	return createPred(PTACC, NULL, createArg(lower, upper), body);
}

pred* createAWC(term* lower, pred* body, term* upper) {
	return createPred(PTAWC, NULL, createArg(lower, upper), body);
}

pred* createWlt(pred* p, term* t) {
	p->weight = t;
	p->vars = appendVars(p->vars, t->vars);
	p->cvar = appendCvar(p->cvar, t->cvar);
	return p;
}

pred* createClt(pred* head, pred* body) {
	head->body = body;
	head->vars = appendVars(head->vars, body->vars);
	head->cvar = appendCvar(head->cvar, body->cvar);
	return head;
}

pred* createCpr(char* string, term* arg1, term* arg2) {
	return createPred(PTCPR, string, createArg(arg1, arg2), NULL);
}

pred* createAss(char* string, term *args) {
	return createPred(PTCPR, strdup("="), createArg(createVar(string), args), NULL);
}

pred* createAtm(char* string, term *args) {
	return createPred(PTREG, string, args, NULL);
}

void flip_cpr(pred* p) {
	if (!strcmp(p->string,"<")) {
		free(p->string);
		p->string = strdup(">=");
	} else if (!strcmp(p->string,">")) {
		free(p->string);
		p->string = strdup("<_");
	} else if (!strcmp(p->string,">=")) {
		free(p->string);
		p->string = strdup("<");
	} else if (!strcmp(p->string,"<=")) {
		free(p->string);
		p->string = strdup(">");
	} else if (!strcmp(p->string,"==")) {
		free(p->string);
		p->string = strdup("!=");
	} else if (!strcmp(p->string,"!=")) {
		free(p->string);
		p->string = strdup("==");
	}
}

rule* createRul(pred* head, pred* body) {
	rule *tmp = (rule*)malloc(sizeof(rule));
	if (head && head->type == PTCON) {
		body = createBod(head, body);
		head->sign ^= 1;
		head = NULL;
	}
	if (head && head->type == PTCPR) {
		flip_cpr(head);
		body = createBod(head, body);
		head = NULL;
	}
	tmp->head = head;
	tmp->body = NULL;
	tmp->cons = NULL;
	pred* it = body;
	while (it) {
		if (it->type == PTCON) {
			pred* jt = it->next;
			it->next = tmp->cons;
			tmp->cons = it;
			it = jt;
		} else {
			pred* jt = it->next;
			it->next = tmp->body;
			tmp->body = it;
			it = jt;
		}
	}
	it = tmp->cons;
	while (it) {
		term* vit = it->vars;
		it->vars = NULL;
		while (vit) {
			term* jt = it->vars;
			while (jt) {
				if (!strcmp(jt->string, vit->string))
					break;
				jt = jt->vars;
			}
			if (jt) {		
				vit = vit->vars;
			} else {
				pred* bit = tmp->body;
				while (bit) {
					term* bvit = bit->vars;
					while (bvit) {
						if (!strcmp(bvit->string, vit->string))
							break;
						bvit = bvit->vars;
					}
					if (bvit)
						break;
					bit = bit->next;
				}
				if (bit) {
					term* sit = vit->vars;
					vit->vars = it->vars;
					it->vars = vit;
					vit = sit;
				} else {					
					vit = vit->vars;
				}
			}
		}
		it = it->next;
	}
	num_vars = 0;
	// reverse paper
	it = tmp->body;
	pred* jt = NULL;
	while (it) {
		pred* cit = it->next;
		it->next = jt;
		jt = it;
		it = cit;
	}
	tmp->body = jt;
	return tmp;
}

void print_arg(term* t) {
	while (t) {
		print_term(t);
		t = t->next;
		if (t) printf(", ");
	}
}

void print_cvar(pred* p) {
	if (p->cvar) {
		term* t = p->cvar;
		while (t) {
			printf(", ");
			enc_cval(t);
			printf(", ");
			enc_cdom(t);
			t = t->cvar;
		}
	}
}

void print_body(pred* p) {
	while (p) {
		print_pred(p);
		print_cvar(p);
		p = p->next;
		if (p) printf(", ");
	}
}

void print_pbody(pred* p) {
	while (p) {
		print_pred(p);
		print_cvar(p);
		p = p->pbody;
		if (p) printf(", ");
	}
}

void print_cond(pred* p) {
	while (p) {
		print_psig(p);
		p = p->body;
		if (p)
			printf(", ");
	}
}

void print_set(pred* p) {
	while (p) {
		print_pred(p);
		term* t = p->cvar;
		while (t) {
			printf(" : ");
			enc_cval(t);
			printf(" : ");
			enc_cdom(t);
			t = t->cvar;
		}
		p = p->next;
		if (p) printf(", ");
	}
}

void print_tsig(term* t) {
	printf("%s", t->string);
	if (t->arity) {
		printf("(");
		print_arg(t->args);
		printf(")");
	}
}

void print_term(term* t) {
	switch (t->type) {
		case TTINT:
			printf("%d",t->number);
			break;
		case TTVAR:
			printf("%s", t->string);
			break;
		case TTRNG:
			print_term(t->args);
			printf("..");
			print_term(t->args->next);
			break;
		case TTFUN:
			print_tsig(t);
			break;
		case TTEXP:
			print_term(t->args);
			printf("%s", t->string);
			print_term(t->args->next);
			break;
		case TTNEG:
			printf("-");
			print_term(t->args);
			break;
		case TTPAR:
			printf("(");
			print_term(t->args);
			printf(")");
			break;
		case TTCVR:
			printf("V_%d",t->ident);
			break;
	}
	if (t->alt) {
		printf(" ; ");
		print_term(t->alt);
	}
}

void print_psig(pred* p) {
	if (p->type == PTCPR) {
		print_term(p->args);
		printf(" %s ",p->string);
		print_term(p->args->next);
	} else {
		printf("%s", p->string);
		if (p->arity) {
			printf("(");
			print_arg(p->args);
			printf(")");
		}
	}
}

void print_csig(pred* p) {
	printf("sig(%d",p->ident);
	term* t = p->vars;
	while (t) {
		printf(", ");
		print_term(t);
		t = t->vars;
	}
	printf(")");
}

void print_pred(pred* p) {
	if (p->sign) printf("not ");
	switch (p->type) {
		case PTREG:
			print_psig(p);
			if (p->body) {
				printf(" : ");
				print_pred(p->body);
			}
			if (p->weight) {
				printf(" = ");
				print_term(p->weight);
			}
			return;
		case PTCPR:
			print_term(p->args);
			printf(" %s ",p->string);
			print_term(p->args->next);
			return;
		case PTACC:
			print_term(p->args);
			if (p->string)
				printf(" %s ", p->string);
			printf(" {");
			print_set(p->body);
			printf("} ");
			print_term(p->args->next);
			return;
		case PTAWC:
			print_term(p->args);
			if (p->string)
				printf(" %s ", p->string);
			printf(" [");
			print_set(p->body);
			printf("] ");
			print_term(p->args->next);
			return;
		case PTCON:
			printf("__constraint(");
			print_csig(p);
			printf(")");
			return;
	}
}

void print_rule(rule* r) {
	pred* lit;
	if (r->head) {
		print_pred(r->head);
	}
	if (r->body || r->cons || (r->head && r->head->cvar)) {
		printf(" :- ");
		print_body(r->body);
		if (r->body && r->cons)
			printf(", ");
		print_body(r->cons);
	}
	if (r->head && r->head->cvar) {
		term* t = r->head->cvar;
		while (t) {
			if (r->body || r->cons)
				printf(", ");
			enc_cval(t);
			printf(", ");
			enc_cdom(t);
			t = t->cvar;
			if (!r->body && !r->cons && t)
				printf(", ");
		}
	}
	printf(".\n");
	lit = r->cons;
	while (lit) {
		lit->sign = 0;
		enc_cons(lit, r->body ? (!r->body->sign && r->body->type == PTREG ? r->body : r->body->pbody) : NULL);
		lit = lit->next;
	}
}

void free_term(term* t) {
	if (t) {
		free(t->string);
		free_term(t->args);
		free_term(t->alt);
		free_term(t->next);
		free_pred(t->body);
		free(t);
	}
}

void free_pred(pred* p) {
	if (p) {
		free(p->string);
		free_term(p->args);
		free_term(p->weight);
		free_term(p->lits);
		free_pred(p->body);
		free_pred(p->next);
		free(p);
	}
}

void free_rule(rule* r) {
	if (r) {
		free_pred(r->head);
		free_pred(r->body);
		free_pred(r->cons);
		free(r);
	}
}

int main(int argc, char** argv) {

	/* read args */
	int opt = 0;
	while ((opt = getopt(argc, argv, ":hv:c:s:a")) != -1) {
		switch (opt) {
			case 'h':
				printf("Usage:\n     inca [options]\n\nOptions:\n  -v <arg>\t: constraint variable encoding\n     direct\t\t: direct encoding\n     bounds\t\t: bounds encoding\n     bounds-hybrid\t: hybrid bounds encoding\n     range\t\t: range encoding\n     range-hybrid\t: hybrid range encoding\n\n  -c <arg>\t: built-in constraint encoding\n     support\t\t: 1-support encoding\n     bounds\t\t: bounds encoding\n     range\t\t: range encoding\n\n  -s <arg>\t: maximal size of Hall intervals\n  -h\t\t: show help\n\nDefault commandline:\n     inca -v bounds-hybrid -c bounds\n\n");
				return 0;
			case 'v':
				if (!strcmp(optarg, "direct"))
					v_enc = VDIRECT;
				else if (!strcmp(optarg, "bounds"))
					v_enc = VBOUNDS;
				else if (!strcmp(optarg, "bounds-hybrid"))
					v_enc = VBOUNDS | VDIRECT;
				else if (!strcmp(optarg, "range"))
					v_enc = VRANGE;
				else if (!strcmp(optarg, "range-hybrid"))
					v_enc = VRANGE | VDIRECT;
				else {
					fprintf(stderr, "error: unknown argument %s for option -%c\n", optarg, optopt);
					return 1;
				}
				break;
			case 'c':
				if (!strcmp(optarg, "support"))
					c_enc = CSUPPORT;
				else if (!strcmp(optarg, "bounds"))
					c_enc = CBOUNDS;
				else if (!strcmp(optarg, "range"))
					c_enc = CRANGE;
				else {
					fprintf(stderr, "error: unknown argument %s for option -%c\n", optarg, optopt);
					return 1;
				}
				break;
			case 's':
				hallsize = atoi(optarg);
				break;
			case ':':
				fprintf(stderr, "error: option -%c without argument\n", optopt);
				return 1;
			case '?':
				fprintf(stderr, "error: unknown option -%c\n", optopt);
				return 1;
		}
	}
	if (optind < argc) {
		fprintf(stderr, "error: unknown parameter %s\n", argv[optind]);
		return 1;
	}

	/* set defaults */
	if (!v_enc || !c_enc) {
		if (c_enc) {
			if (c_enc == CSUPPORT)
				v_enc = VDIRECT;
			else
				v_enc = c_enc | VDIRECT;
		} else if (v_enc) {
			c_enc = (v_enc & CBOUNDS) ? CBOUNDS : ((v_enc & CRANGE) ? CRANGE : CSUPPORT);
		} else {
			v_enc = VBHYBRID;
			c_enc = CBOUNDS;
		}
	}

	/* check args */
	if (!(v_enc == c_enc) && !(v_enc == VBHYBRID && c_enc != CRANGE) && !(v_enc == VRHYBRID && c_enc != CBOUNDS)) {
		fprintf(stderr, "error: specified variable- and constraint encoding don't match\n");
		return 1;
	}

	/* parse and return result */
	int res = yyparse();
	if (!res) {
		printf("#hide __constraint(X).\n");
		enc_hide();
	}
	return res;
}

